import uuid

import stringcase


def snake_case_keys(whatever_dict):
    """
    Converts keys in dict to snake case.
    """
    return {stringcase.snakecase(k): v for k, v in whatever_dict.items()}


class EntityObject(object):
    """
    Converts json returned form server into a proper entity object.
    """

    def update(self, update_attrs):
        entity = self.__client.update_by_id(self.uuid, update_attrs)
        self.__dict__ = entity.__dict__

    def delete(self):
        deleted = self.__client.delete_by_id(self.uuid)
        if deleted:
            self.__dict__ = {"uuid": self.uuid, "deleted": deleted}

    def set_alias_for(self, alias_for):
        entity = self.__client.alias_by_id(self.uuid, alias_for.uuid)
        self.__dict__ = entity.__dict__

    def remove_alias_for(self):
        entity = self.__client.alias_by_id(self.uuid, None)
        self.__dict__ = entity.__dict__

    def set_superseded_by(self, superseded_by):
        entity = self.__client.supersede_by_id(self.uuid, superseded_by.uuid)
        self.__dict__ = entity.__dict__

    def remove_superseded_by(self):
        entity = self.__client.supersede_by_id(self.uuid, None)
        self.__dict__ = entity.__dict__

    def __init__(self, response, client=None):
        entity = response.pop("entity", {})
        attributes = entity.pop("attributes", {})

        self.__dict__ = {
            **snake_case_keys(response),
            **snake_case_keys(entity),
            **snake_case_keys(attributes),
        }

        # Manually decode UUIDs, which are strings when decoded from json
        self.uuid = uuid.UUID(self.uuid)
        if self.superseded_by:
            self.superseded_by = uuid.UUID(self.superseded_by)
        if self.alias_for:
            self.alias_for = uuid.UUID(self.alias_for)

        self.__client = client

    def __repr__(self):
        return f"<Crosswalk Entity: {str(self.uuid)[:6]}...>"



from slugify import slugify

from crosswalk_client.exceptions import MalformedDomain, MissingDomain
from crosswalk_client.objects.domain import DomainObject


def validate_required_domain_arg(function):
    """
    Validates a domain is passed. Converts DomainObject instances to slug.
    """

    def wrapper(*args, **kwargs):
        list_args = list(args)
        try:
            domain = list_args[1]
        except IndexError:
            raise MissingDomain("Must pass a domain")
        if isinstance(domain, DomainObject):
            list_args[1] = domain.slug
            return function(*list_args, **kwargs)
        if not isinstance(domain, str):
            raise MalformedDomain("Domain must be a string or domain instance")
        if slugify(domain) != domain:
            raise MalformedDomain("Domain string should be a slug")
        return function(*args, **kwargs)

    return wrapper



# flake8: noqa
from .block_attrs_kwarg import validate_block_attrs_kwarg
from .create_attrs_kwarg import validate_create_attrs_kwarg
from .domain_kwarg import validate_domain_kwarg
from .required_block_attrs_arg import validate_required_block_attrs_arg
from .required_query_arg import validate_required_query_arg
from .required_update_attrs_arg import validate_required_update_attrs_arg
from .required_uuid_arg import validate_required_uuid_arg
from .target_uuid_arg import validate_target_uuid_arg
from .threshold_kwarg import validate_threshold_kwarg
from .update_attrs_kwarg import validate_update_attrs_kwarg



from urllib.parse import urljoin

import requests

from crosswalk_client.encoder import encode
from crosswalk_client.exceptions import (
    BadRequest,
    BadResponse,
    UnspecificDeleteRequestError,
)
from crosswalk_client.validators.entity import (
    validate_domain_kwarg,
    validate_required_block_attrs_arg,
)


class DeleteMatch(object):
    @validate_required_block_attrs_arg
    @validate_domain_kwarg
    def delete_match(self, block_attrs, domain=None):
        if not isinstance(block_attrs, dict):
            raise BadRequest("block_attrs must be a dictionary of attributes.")
        if domain is None:
            domain = self.domain
        data = block_attrs
        response = requests.post(
            urljoin(
                self.service_address,
                "domains/{}/entities/delete-match/".format(domain),
            ),
            headers=self.headers,
            data=encode(data),
        )
        if response.status_code == 403:
            raise UnspecificDeleteRequestError(
                "Delete request matched more than one record."
            )
        if response.status_code != 204:
            raise BadResponse(
                "The service responded with a {}: {}".format(
                    response.status_code, response.content
                )
            )
        return response.status_code == 204



from crosswalk_client import Client


def test_setup(token, service):
    client = Client(token, service)
    client.create_domain("presidents")
    client.create_domain("politicians")
    client.bulk_create([{"name": "George W. Bush"}], domain="politicians")
    client.bulk_create([{"name": "George W. Bush"}], domain="presidents")
    client.bulk_create([{"name": "George Bush, Jr."}], domain="presidents")


def test_entity_update(token, service):
    client = Client(token, service, domain="presidents")
    entity = client.best_match({"name": "George W. Bush"})
    entity.update({"number": 43})
    assert entity.number == 43


def test_entity_set_alias_for(token, service):
    client = Client(token, service, domain="presidents")
    entity = client.best_match({"name": "George W. Bush"})
    alias = client.best_match({"name": "George Bush, Jr."})
    alias.set_alias_for(entity)
    assert alias.alias_for == entity.uuid


def test_entity_remove_alias_for(token, service):
    client = Client(token, service, domain="presidents")
    entity = client.best_match(
        {"name": "George Bush, Jr."}, return_canonical=False
    )
    assert entity.name == "George Bush, Jr."
    entity.remove_alias_for()
    assert entity.alias_for is None


def test_entity_set_superseded_by(token, service):
    client = Client(token, service)
    entity = client.best_match({"name": "George W. Bush"}, domain="presidents")
    superseded = client.best_match(
        {"name": "George W. Bush"}, domain="politicians"
    )
    superseded.set_superseded_by(entity)
    assert superseded.superseded_by == entity.uuid


def test_entity_remove_superseded_by(token, service):
    client = Client(token, service, domain="politicians")
    entity = client.best_match({"name": "George W. Bush"})
    entity.remove_superseded_by()
    assert entity.superseded_by is None


def test_entity_delete(token, service):
    client = Client(token, service, domain="presidents")
    entity = client.best_match({"name": "George W. Bush"})
    entity.delete()
    assert entity.deleted is True

    entity = client.best_match(
        {"name": "George Bush, Jr."}, return_canonical=False
    )
    entity.delete()
    assert entity.deleted is True

    entity = client.best_match(
        {"name": "George W. Bush"}, domain="politicians"
    )
    entity.delete()
    assert entity.deleted is True


def test_cleanup(token, service):
    client = Client(token, service)
    client.delete_domain("presidents")
    client.delete_domain("politicians")



from urllib.parse import urljoin

import requests

from crosswalk_client.exceptions import BadResponse, ProtectedDomainError
from crosswalk_client.validators.domain import validate_required_domain_arg


class DeleteDomain(object):
    @validate_required_domain_arg
    def delete_domain(self, domain):
        response = requests.delete(
            urljoin(self.service_address, f"domains/{domain}/"),
            headers=self.headers,
        )
        if response.status_code == 500:
            raise ProtectedDomainError(
                "Could not delete domain. It may have protected entites."
            )
        if response.status_code != 204:
            raise BadResponse(
                "The service responded with a {}: {}".format(
                    response.status_code, response.content
                )
            )
        return True



from urllib.parse import urljoin

import requests

from crosswalk_client.exceptions import BadResponse
from crosswalk_client.objects.domain import DomainObject
from crosswalk_client.validators.domain import validate_parent_domain_kwarg


class GetDomains(object):
    @validate_parent_domain_kwarg
    def get_domains(self, parent=None):
        params = {}
        if parent:
            params["parent"] = parent
        response = requests.get(
            urljoin(self.service_address, "domains/"),
            headers=self.headers,
            params=params,
        )
        if response.status_code != requests.codes.ok:
            raise BadResponse(
                "The service responded with a {}: {}".format(
                    response.status_code, response.content
                )
            )
        domains = response.json()
        return [DomainObject(domain, client=self) for domain in domains]



import pytest
from crosswalk_client import Client
from crosswalk_client.exceptions import (
    MalformedDomain,
    MissingDomain,
    ProtectedDomainError,
)


def test_create_domain(token, service):
    client = Client(token, service)
    states = client.create_domain("states")
    assert states.name == "states"

    counties = client.create_domain("counties", parent="states")
    assert counties.name == "counties" and counties.parent == "states"

    cities = client.create_domain("cities", parent=states)
    assert cities.name == "cities" and cities.parent == "states"

    with pytest.raises(MalformedDomain):
        client.create_domain(1)

    with pytest.raises(MalformedDomain):
        client.create_domain("townships", parent=1)

    townships = client.create_domain("townships", parent=states)
    assert townships.name == "townships" and townships.parent == "states"


def test_create_domain_exceptions(token, service):
    client = Client(token, service)
    with pytest.raises(MissingDomain):
        client.create_domain()
    with pytest.raises(MalformedDomain):
        client.create_domain("countries", parent="a parent name")
    with pytest.raises(MalformedDomain):
        client.create_domain("countries", parent=1)


def test_update_domain(token, service):
    client = Client(token, service)
    # Client passed slug
    domain = client.update_domain(
        "counties", {"name": "kounties", "parent": None}
    )
    assert domain.name == "kounties" and domain.parent is None
    # Client passed instance
    domain = client.update_domain(domain, {"name": "counties", "parent": None})
    assert domain.name == "counties"
    # Instance method
    domain.update({"name": "kounties"})
    assert domain.name == "kounties"
    # Exception
    with pytest.raises(MalformedDomain):
        client.update_domain(1, {"name": "counties"})


def test_get_domain(token, service):
    client = Client(token, service)
    domain = client.get_domain("states")
    assert domain.slug == "states"

    with pytest.raises(MalformedDomain):
        client.get_domain(1)


def test_get_domains(token, service):
    client = Client(token, service)
    domains = client.get_domains()
    states = domains[0]
    assert states.slug == "states"

    domains = client.get_domains(parent=states)
    assert len(domains) == 2

    domains = client.get_domains(parent="states")
    assert len(domains) == 2


def test_get_domains_exceptions(token, service):
    client = Client(token, service)
    with pytest.raises(MalformedDomain):
        client.get_domains(parent=1)
    with pytest.raises(MalformedDomain):
        client.get_domains(parent="U.S. states")


def test_delete_protected_domain_error(token, service):
    client = Client(token, service)
    client.bulk_create([{"name": "Kansas"}], domain="states")
    with pytest.raises(ProtectedDomainError):
        client.delete_domain("states")


def test_delete_domain(token, service):
    client = Client(token, service)
    # Client passed slug
    response = client.delete_domain("cities")
    assert response is True
    # Client passed instance
    domain = client.get_domain("counties")
    response = client.delete_domain(domain)
    assert response is True
    # Instance method
    domain = client.get_domain("townships")
    domain.delete()
    assert domain.deleted is True


def test_cleanup(token, service):
    client = Client(token, service, domain="states")
    client.delete_match({"name": "Kansas"})
    client.delete_domain("states")



import pytest
from crosswalk_client import Client
from crosswalk_client.exceptions import (
    MalformedBlockAttributes,
    MalformedDomain,
    MalformedQuery,
    MalformedThreshold,
    MalformedUUID,
    MissingDomain,
)


def test_validate_query_not_a_dictionary(token, service):
    client = Client(token, service, domain="states")
    with pytest.raises(MalformedQuery):
        client.alias_or_create("test")


def test_validate_query_too_many_query_fields(token, service):
    client = Client(token, service, domain="states")
    with pytest.raises(MalformedQuery):
        client.alias_or_create({"name": "Alderaan", "galaxy": "Far, far away"})


def test_validate_query_query_not_a_string(token, service):
    client = Client(token, service, domain="states")
    with pytest.raises(MalformedQuery):
        client.alias_or_create({"name": 1})


def test_validate_domain_missing(token, service):
    client = Client(token, service)
    with pytest.raises(MissingDomain):
        client.get_entities()


def test_validate_domain_not_a_slug(token, service):
    client = Client(token, service)
    with pytest.raises(MalformedDomain):
        client.get_entities(domain="NotASlug")


def test_validate_domain_not_a_string(token, service):
    client = Client(token, service)
    with pytest.raises(MalformedDomain):
        client.get_entities(domain=123)


def validate_block_attrs_not_a_dictionary(token, service):
    client = Client(token, service)
    with pytest.raises(MalformedBlockAttributes):
        client.best_match({"name": "entity"}, block_attrs="not a dictionary")


def validate_block_attrs_nested(token, service):
    client = Client(token, service)
    with pytest.raises(MalformedBlockAttributes):
        client.best_match(
            {"name": "entity"}, block_attrs={"nested": {"too": "deep"}}
        )


def test_validate_uuid_malformed(token, service):
    client = Client(token, service)
    with pytest.raises(MalformedUUID):
        client.delete_by_id("2bc1c94f 0deb-43e9-92a1-4775189ec9f8")


def test_validate_threshold_not_an_integer(token, service):
    client = Client(token, service)
    with pytest.raises(MalformedThreshold):
        client.best_match_or_create(
            {"name": "Xanadu"}, domain="domain", threshold="bad"
        )


def test_validate_threshold_out_of_range(token, service):
    client = Client(token, service)
    with pytest.raises(MalformedThreshold):
        client.best_match_or_create(
            {"name": "Xanadu"}, domain="domain", threshold=200
        )



from crosswalk_client.exceptions import MalformedCreateAttributes


def validate_create_attrs_kwarg(function):
    def wrapper(*args, **kwargs):
        create_attrs = kwargs.get("create_attrs", {}).copy()
        if not isinstance(create_attrs, dict):
            raise MalformedCreateAttributes(
                "Create attributes must be a dictionary"
            )
        for key in create_attrs:
            if isinstance(create_attrs[key], dict):
                raise MalformedCreateAttributes(
                    "Nested attributes are not allowed in create attributes"
                )
        return function(*args, **kwargs)

    return wrapper



# flake8: noqa
from .required_domain_arg import validate_required_domain_arg
from .required_domain_kwarg import validate_required_domain_kwarg
from .required_domain_string_arg import validate_required_domain_string_arg
from .parent_domain_kwarg import validate_parent_domain_kwarg



from urllib.parse import urljoin

import requests

from crosswalk_client.encoder import encode
from crosswalk_client.exceptions import BadResponse
from crosswalk_client.objects.entity import EntityObject
from crosswalk_client.validators.entity import (
    validate_required_update_attrs_arg,
    validate_required_uuid_arg,
)


class UpdateById(object):
    """
    Update an entity.
    """

    @validate_required_uuid_arg
    @validate_required_update_attrs_arg
    def update_by_id(self, uuid, update_attrs):
        response = requests.patch(
            urljoin(self.service_address, "entities/{}/".format(uuid)),
            headers=self.headers,
            data=encode({"attributes": update_attrs}),
        )
        if response.status_code != requests.codes.ok:
            raise BadResponse(
                "The service responded with a {}: {}".format(
                    response.status_code, response.content
                )
            )
        return EntityObject({"entity": response.json()}, client=self)



import pytest
from crosswalk_client import Client
from crosswalk_client.exceptions import (
    ConfigError,
    MalformedDomain,
    MalformedScorer,
    MalformedThreshold,
    MissingDomain,
    MissingScorer,
    MissingThreshold,
)


def test_misconfigured_config(token, service):
    with pytest.raises(ConfigError):
        Client("NOTAVALIDTOKEN", service)


def test_properly_configured_config(token, service):
    client = Client(token, service)
    response = client.client_check()
    assert response is True


def test_set_threshold(token, service):
    client = Client(token, service)
    client.set_threshold(99)
    assert client.threshold == 99


def test_set_threshold_exceptions(token, service):
    client = Client(token, service)
    with pytest.raises(MissingThreshold):
        client.set_threshold()
    with pytest.raises(MalformedThreshold):
        client.set_threshold(200)
    with pytest.raises(MalformedThreshold):
        client.set_threshold("asd")


def test_set_domain(token, service):
    client = Client(token, service)
    client.set_domain("sharks")
    assert client.domain == "sharks"

    domain = client.create_domain("whales")
    client.set_domain(domain)
    assert client.domain == "whales"

    domain.delete()


def test_set_domain_exceptions(token, service):
    client = Client(token, service)
    with pytest.raises(MissingDomain):
        client.set_domain()
    with pytest.raises(MalformedDomain):
        client.set_domain("U.S. states")


def test_set_scorer(token, service):
    client = Client(token, service)
    client.set_scorer("fuzzywuzzy.partial_ratio_process")
    assert client.scorer == "fuzzywuzzy.partial_ratio_process"


def test_set_scorer_exceptions(token, service):
    client = Client(token, service)
    with pytest.raises(MissingScorer):
        client.set_scorer()
    with pytest.raises(MalformedScorer):
        client.set_scorer("random.scorer")



from urllib.parse import urljoin

import requests

from crosswalk_client.exceptions import BadResponse
from crosswalk_client.objects.entity import EntityObject
from crosswalk_client.validators.entity import validate_required_uuid_arg


class GetEntity(object):
    @validate_required_uuid_arg
    def get_entity(self, uuid):
        response = requests.get(
            urljoin(self.service_address, "entities/{}/".format(uuid)),
            headers=self.headers,
        )
        if response.status_code != requests.codes.ok:
            raise BadResponse(
                "The service responded with a {}: {}".format(
                    response.status_code, response.content
                )
            )
        entity = response.json()
        return EntityObject({"entity": entity}, client=self)
